import { isEqual } from '../isEqual/index.js';
const mockSymbol = Symbol('mock');
// biome-ignore lint/suspicious/noExplicitAny: any is required here
export function mockFn(defaultImplementation) {
    let defaultSpec = {
        type: 'not-ready',
    };
    let oneTimeOverrides = [];
    let parameterOverrides = [];
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    function mock(...args) {
        for (const override of parameterOverrides) {
            if (isEqual(args, override.args)) {
                parameterOverrides.splice(parameterOverrides.indexOf(override), 1);
                return runSpec(override.spec, args);
            }
        }
        const current = oneTimeOverrides.shift() ?? defaultSpec;
        return runSpec(current, args);
    }
    mock[mockSymbol] = true;
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.calls = [];
    mock.isExhausted = () => oneTimeOverrides.length === 0 && parameterOverrides.length === 0;
    mock.getOneTimeOverridesLength = () => oneTimeOverrides.length;
    mock.getParameterOverridesLength = () => parameterOverrides.length;
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    function runSpec(spec, args) {
        switch (spec.type) {
            case 'return': {
                mock.calls.push({ args, result: { type: 'return', value: spec.value } });
                return spec.value;
            }
            case 'lazy-return': {
                const value = spec.value();
                mock.calls.push({ args, result: { type: 'return', value } });
                return value;
            }
            case 'throw': {
                mock.calls.push({ args, result: { type: 'throw', error: spec.error } });
                throw spec.error;
            }
            case 'exec': {
                try {
                    const value = spec.implementation(...args);
                    mock.calls.push({ args, result: { type: 'return', value } });
                    return value;
                }
                catch (error) {
                    mock.calls.push({ args, result: { type: 'throw', error } });
                    throw error;
                }
            }
            case 'not-ready': {
                if (parameterOverrides.length > 0) {
                    throw new Error('The mock function was called with arguments that do not match any of the parameter overrides and no default behavior has been provided.');
                }
                throw new TypeError('The mock function was called but no default behavior has been provided.');
            }
            default: {
                throw new Error('Unreachable case');
            }
        }
    }
    mock.reset = () => {
        defaultSpec = { type: 'not-ready' };
        oneTimeOverrides = [];
        parameterOverrides = [];
        mock.clear();
        if (defaultImplementation) {
            mock.executes(defaultImplementation);
        }
    };
    mock.clear = () => {
        mock.calls = [];
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.returns = (value) => {
        defaultSpec = { type: 'return', value };
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.returnsOnce = (value) => {
        oneTimeOverrides.push({ type: 'return', value });
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.throws = (error) => {
        defaultSpec = { type: 'throw', error };
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.throwsOnce = (error) => {
        oneTimeOverrides.push({ type: 'throw', error });
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.executes = (implementation) => {
        defaultSpec = { type: 'exec', implementation };
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.executesOnce = (implementation) => {
        oneTimeOverrides.push({ type: 'exec', implementation });
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.resolvesTo = (value) => {
        defaultSpec = { type: 'return', value: Promise.resolve(value) };
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.resolvesToOnce = (value) => {
        oneTimeOverrides.push({ type: 'return', value: Promise.resolve(value) });
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.rejectsWith = (error) => {
        defaultSpec = { type: 'lazy-return', value: () => Promise.reject(error) };
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.rejectsWithOnce = (error) => {
        oneTimeOverrides.push({
            type: 'lazy-return',
            value: () => Promise.reject(error),
        });
        return mock;
    };
    // biome-ignore lint/suspicious/noExplicitAny: any is required here
    mock.given = (...args) => ({
        // biome-ignore lint/suspicious/noExplicitAny: any is required here
        returnsOnce(value) {
            parameterOverrides.push({ args, spec: { type: 'return', value } });
            return mock;
        },
        // biome-ignore lint/suspicious/noExplicitAny: any is required here
        throwsOnce(error) {
            parameterOverrides.push({ args, spec: { type: 'throw', error } });
            return mock;
        },
        // biome-ignore lint/suspicious/noExplicitAny: any is required here
        executesOnce(implementation) {
            parameterOverrides.push({
                args,
                spec: { type: 'exec', implementation },
            });
            return mock;
        },
        // biome-ignore lint/suspicious/noExplicitAny: any is required here
        resolvesToOnce(value) {
            parameterOverrides.push({
                args,
                spec: { type: 'return', value: Promise.resolve(value) },
            });
            return mock;
        },
        // biome-ignore lint/suspicious/noExplicitAny: any is required here
        rejectsWithOnce(error) {
            parameterOverrides.push({
                args,
                spec: { type: 'lazy-return', value: () => Promise.reject(error) },
            });
            return mock;
        },
    });
    if (defaultImplementation) {
        mock.executes(defaultImplementation);
    }
    return mock;
}
// biome-ignore lint/suspicious/noExplicitAny: any is required here
export function isMockFn(value) {
    return typeof value === 'function' && mockSymbol in value;
}
//# sourceMappingURL=mockFn.js.map